@@ -4,11 +4,27 @@
 
 require 'puppet'
 require 'timeout'
+require 'puppet/rails'
 require 'puppet/parser/parser'
 require 'puppet/parser/scope'
 
 class Puppet::Parser::Interpreter
     include Puppet::Util
+    
+    Puppet.setdefaults(:puppet,
+        :casesensitive => [false,
+            "Whether matching in case statements and selectors
+            should be case-sensitive.  Case insensitivity is
+            handled by downcasing all values before comparison."],
+        :external_nodes => ["none",
+            "An external command that can produce node information.  The
+            first line of output must be either the parent node or blank,
+            and if there is a second line of output it should be a list of
+            whitespace-separated classes to include on that node.  This command
+            makes it straightforward to store your node mapping information
+            in other data sources like databases.
+            
+            For unknown nodes, the commands should exit with an exit code of 1."])
 
     Puppet.setdefaults("ldap",
         :ldapnodes => [false,
@@ -81,25 +97,64 @@ class Puppet::Parser::Interpreter
         return @ldap
     end
 
+    # Make sure we don't have any remaining collections that specifically
+    # look for resources, because we want to consider those to be
+    # parse errors.
+    def check_resource_collections(scope)
+        remaining = []
+        scope.collections.each do |coll|
+            if r = coll.resources
+                if r.is_a?(Array)
+                    remaining += r
+                else
+                    remaining << r
+                end
+            end
+        end
+        unless remaining.empty?
+            raise Puppet::ParseError, "Failed to find virtual resources %s" %
+                remaining.join(', ')
+        end
+    end
+
     def clear
         initparsevars
     end
 
-    # Iteratively evaluate all of the objects.  This finds all fo the
-    # objects that represent definitions and evaluates the definitions appropriately.
+    # Iteratively evaluate all of the objects.  This finds all of the objects
+    # that represent definitions and evaluates the definitions appropriately.
     # It also adds defaults and overrides as appropriate.
     def evaliterate(scope)
         count = 0
-        begin
-            timeout 300 do
-                while ary = scope.unevaluated
-                    ary.each do |resource|
-                        resource.evaluate
+        loop do
+            count += 1
+            done = true
+            # First perform collections, so we can collect defined types.
+            if coll = scope.collections and ! coll.empty?
+                exceptwrap do
+                    coll.each do |c|
+                        # Only keep the loop going if we actually successfully
+                        # collected something.
+                        if o = c.evaluate
+                            done = false
+                        end
                     end
                 end
             end
-        rescue Timeout::Error
-            raise Puppet::DevError, "Got a timeout trying to evaluate all definitions"
+            
+            # Then evaluate any defined types.
+            if ary = scope.unevaluated
+                ary.each do |resource|
+                    resource.evaluate
+                end
+                # If we evaluated, then loop through again.
+                done = false
+            end
+            break if done
+            
+            if count > 1000
+                raise Puppet::ParseError, "Got 1000 class levels, which is unsupported"
+            end
         end
     end
 
@@ -142,7 +197,7 @@ class Puppet::Parser::Interpreter
         scope.name = "top"
         scope.type = "main"
 
-        scope.host = facts["hostname"] || Facter.value("hostname")
+        scope.host = client || facts["hostname"] || Facter.value(:hostname)
 
         classes = @classes.dup
 
@@ -175,13 +230,6 @@ class Puppet::Parser::Interpreter
         # Now make sure we fail if there's anything left to do
         failonleftovers(scope)
 
-        # Now perform the collections
-        exceptwrap do
-            scope.collections.each do |coll|
-                coll.evaluate
-            end
-        end
-
         # Now finish everything.  This recursively calls finish on the
         # contained scopes and resources.
         scope.finish
@@ -191,7 +239,7 @@ class Puppet::Parser::Interpreter
         if Puppet[:storeconfigs]
             args = {
                 :resources => scope.resources,
-                :name => client,
+                :name => scope.host,
                 :facts => facts
             }
             unless scope.classlist.empty?
@@ -216,14 +264,15 @@ class Puppet::Parser::Interpreter
     # Fail if there any overrides left to perform.
     def failonleftovers(scope)
         overrides = scope.overrides
-        if overrides.empty?
-            return nil
-        else
+        unless overrides.empty?
             fail Puppet::ParseError,
                 "Could not find object(s) %s" % overrides.collect { |o|
                     o.ref
                 }.join(", ")
         end
+
+        # Now check that there aren't any extra resource collections.
+        check_resource_collections(scope)
     end
 
     # Find a class definition, relative to the current namespace.
@@ -238,6 +287,8 @@ class Puppet::Parser::Interpreter
 
     # The recursive method used to actually look these objects up.
     def fqfind(namespace, name, table)
+        namespace = namespace.downcase
+        name = name.downcase
         if name =~ /^::/ or namespace == ""
             return table[name.sub(/^::/, '')]
         end
@@ -263,7 +314,7 @@ class Puppet::Parser::Interpreter
     end
 
     # Create a new node, just from a list of names, classes, and an optional parent.
-    def gennode(name, hash)
+    def gennode(name, hash, source = nil)
         facts = hash[:facts]
         classes = hash[:classes]
         parent = hash[:parentnode]
@@ -297,7 +348,14 @@ class Puppet::Parser::Interpreter
         end
 
         # Create the node
-        return @parser.ast(AST::Node, arghash)
+        if source
+            arghash[:file] = source
+        else
+            arghash[:file] = nil
+        end
+        arghash[:line] = nil
+        node = @parser.ast(AST::Node, arghash)
+        return node
     end
 
     # create our interpreter
@@ -338,6 +396,10 @@ class Puppet::Parser::Interpreter
         unless @nodesources.include?(:code)
             @nodesources << :code
         end
+        
+        unless Puppet[:external_nodes] == "none"
+            @nodesources << :external
+        end
 
         @setup = false
 
@@ -358,8 +420,12 @@ class Puppet::Parser::Interpreter
         end
 
         # The class won't always be defined during testing.
-        if Puppet[:storeconfigs] and defined? ActiveRecord::Base
-            Puppet::Rails.init
+        if Puppet[:storeconfigs] 
+            if Puppet.features.rails?
+                Puppet::Rails.init
+            else
+                raise Puppet::Error, "Rails is missing; cannot store configurations"
+            end
         end
 
         @files = []
@@ -389,10 +455,6 @@ class Puppet::Parser::Interpreter
             end
         end
 
-        if node =~ /\./
-            node = node.sub(/\..+/, '')
-        end
-
         filter = Puppet[:ldapstring]
         attrs = Puppet[:ldapattrs].split("\s*,\s*")
         sattrs = attrs.dup
@@ -467,6 +529,7 @@ class Puppet::Parser::Interpreter
 
     # Create a new class, or merge with an existing class.
     def newclass(fqname, options = {})
+        fqname = fqname.downcase
         if @definetable.include?(fqname)
             raise Puppet::ParseError, "Cannot redefine class %s as a definition" %
                 fqname
@@ -478,8 +541,7 @@ class Puppet::Parser::Interpreter
         if other = @classtable[fqname]
             # Make sure the parents match
             if parent and other.parentclass and (parent != other.parentclass)
-                @parser.error @parser.addcontext("Class %s is already defined" % fqname) +
-                    " with parent %s" % [fqname, other.parentclass]
+                @parser.error("Class %s is already defined at %s:%s; cannot redefine" % [fqname, other.file, other.line])
             end
 
             # This might be dangerous...
@@ -505,7 +567,11 @@ class Puppet::Parser::Interpreter
         else
             # Define it anew.
             ns, name = namesplit(fqname)
-            args = {:type => name, :namespace => ns, :fqname => fqname, :interp => self}
+
+            # Note we're doing something somewhat weird here -- we're setting
+            # the class's namespace to its fully qualified name.  This means
+            # anything inside that class starts looking in that namespace first.
+            args = {:type => name, :namespace => fqname, :fqname => fqname, :interp => self}
             args[:code] = code if code
             args[:parentclass] = parent if parent
             @classtable[fqname] = @parser.ast AST::HostClass, args
@@ -516,16 +582,14 @@ class Puppet::Parser::Interpreter
 
     # Create a new definition.
     def newdefine(fqname, options = {})
+        fqname = fqname.downcase
         if @classtable.include?(fqname)
             raise Puppet::ParseError, "Cannot redefine class %s as a definition" %
                 fqname
         end
         # Make sure our definition doesn't already exist
         if other = @definetable[fqname]
-            @parser.error @parser.addcontext(
-                "%s is already defined at line %s" % [fqname, other.line],
-                other
-            )
+            @parser.error("%s is already defined at %s:%s; cannot redefine" % [fqname, other.file, other.line])
         end
 
         ns, name = namesplit(fqname)
@@ -549,8 +613,9 @@ class Puppet::Parser::Interpreter
     def newnode(names, options = {})
         names = [names] unless names.instance_of?(Array)
         names.collect do |name|
+            name = name.to_s.downcase
             if other = @nodetable[name]
-                @parser.error @parser.addcontext("Node %s is already defined" % [other.name], other)
+                @parser.error("Node %s is already defined at %s:%s; cannot redefine" % [other.name, other.file, other.line])
             end
             name = name.to_s if name.is_a?(Symbol)
             args = {
@@ -574,8 +639,8 @@ class Puppet::Parser::Interpreter
     # reparsed.
     def newfile(*files)
         files.each do |file|
-            unless file.is_a? Puppet::LoadedFile
-                file = Puppet::LoadedFile.new(file)
+            unless file.is_a? Puppet::Util::LoadedFile
+                file = Puppet::Util::LoadedFile.new(file)
             end
             @files << file
         end
@@ -583,6 +648,7 @@ class Puppet::Parser::Interpreter
 
     # Search for our node in the various locations.
     def nodesearch(*nodes)
+        nodes = nodes.collect { |n| n.to_s.downcase }
         # At this point, stop at the first source that defines
         # the node
         @nodesources.each do |source|
@@ -617,6 +683,45 @@ class Puppet::Parser::Interpreter
     def nodesearch_code(name)
         @nodetable[name]
     end
+    
+    # Look for external node definitions.
+    def nodesearch_external(name)
+        return nil unless Puppet[:external_nodes] != "none"
+        
+        begin
+            output = Puppet::Util.execute([Puppet[:external_nodes], name])
+        rescue Puppet::ExecutionFailure => detail
+            if $?.exitstatus == 1
+                return nil
+            else
+                Puppet.err "Could not retrieve external node information for %s: %s" % [name, detail]
+            end
+            return nil
+        end
+        
+        if output =~ /\A\s+\Z/ # all whitespace
+            puts "empty response for %s" % name
+            return nil
+        end
+        
+        lines = output.split("\n")
+        
+        args = {}
+        parent = lines[0].gsub(/\s+/, '')
+        args[:parentnode] = parent unless parent == ""
+        
+        if lines[1]
+            classes = lines[1].sub(/^\s+/,'').sub(/\s+$/,'').split(/\s+/)
+            args[:classes] = classes unless classes.empty?
+        end
+        
+        if args.empty?
+            Puppet.warning "Somehow got a node with no information"
+            return nil
+        else
+            return gennode(name, args, Puppet[:external_nodes])
+        end
+    end
 
     # Look for our node in ldap.
     def nodesearch_ldap(node)
@@ -625,7 +730,7 @@ class Puppet::Parser::Interpreter
             args = {}
             args[:classes] = classes if classes
             args[:parentnode] = parent if parent
-            return gennode(node, args)
+            return gennode(node, args, "ldap")
         else
             return nil
         end
@@ -663,15 +768,14 @@ class Puppet::Parser::Interpreter
     # Connect to the LDAP Server
     def setup_ldap
         self.class.ldap = nil
-        begin
-            require 'ldap'
-        rescue LoadError
+        unless Puppet.features.ldap?
             Puppet.notice(
                 "Could not set up LDAP Connection: Missing ruby/ldap libraries"
             )
             @ldap = nil
             return
         end
+
         begin
             @ldap = self.class.ldap()
         rescue => detail
@@ -683,10 +787,6 @@ class Puppet::Parser::Interpreter
         return @scope
     end
 
-    # Iteratively make sure that every object in the scope tree is translated.
-    def translate(scope)
-    end
-
     private
 
     # Check whether any of our files have changed.
@@ -753,15 +853,14 @@ class Puppet::Parser::Interpreter
 
     # Store the configs into the database.
     def storeconfigs(hash)
-        unless defined? ActiveRecord
-            require 'puppet/rails'
-            unless defined? ActiveRecord
-                raise LoadError,
-                    "storeconfigs is enabled but rails is unavailable"
-            end
+        unless Puppet.features.rails?
+            raise Puppet::Error,
+                "storeconfigs is enabled but rails is unavailable"
         end
 
-        Puppet::Rails.init
+        unless ActiveRecord::Base.connected?
+            Puppet::Rails.connect
+        end
 
         # Fork the storage, since we don't need the client waiting
         # on that.  How do I avoid this duplication?
@@ -777,19 +876,21 @@ class Puppet::Parser::Interpreter
                 end
             }
         else
-            # We store all of the objects, even the collectable ones
-            benchmark(:info, "Stored configuration for #{hash[:name]}") do
-                Puppet::Rails::Host.transaction do
-                    Puppet::Rails::Host.store(hash)
+            begin
+                # We store all of the objects, even the collectable ones
+                benchmark(:info, "Stored configuration for #{hash[:name]}") do
+                    Puppet::Rails::Host.transaction do
+                        Puppet::Rails::Host.store(hash)
+                    end
                 end
+            rescue => detail
+                if Puppet[:trace]
+                    puts detail.backtrace
+                end
+                Puppet.err "Could not store configs: %s" % detail.to_s
             end
         end
-
-        # Now that we've stored everything, we need to strip out
-        # the collectable objects so that they are not sent on
-        # to the host
-        #hash[:objects].collectstrip!
     end
 end
 
-# $Id: interpreter.rb 1805 2006-10-18 00:22:45Z luke $
+# $Id: interpreter.rb 2432 2007-04-30 15:38:06Z luke $
