@@ -51,7 +51,7 @@ module Puppet
                 %r{([A-Z][-\w]*::)+[A-Z][-\w]*} => :CLASSREF,
                 %r{[A-Z][-\w]*} => :TYPE,
                 %r{[0-9]+} => :NUMBER,
-                %r{\$\w+} => :VARIABLE
+                %r{\$(\w*::)*\w+} => :VARIABLE
             }
 
             @@keywords = {
@@ -81,6 +81,8 @@ module Puppet
                 array = []
 
                 self.scan { |token,str|
+                    # Ignore any definition nesting problems
+                    @indefine = false
                     #Puppet.debug("got token '%s' => '%s'" % [token,str])
                     if token.nil?
                         return array
@@ -252,12 +254,12 @@ module Puppet
 
                     if @lasttoken == :DEFINE
                         if indefine?
+                            msg = "Cannot nest definition %s inside %s" % [value, @indefine]
                             self.indefine = false
-                            raise Puppet::ParseError,
-                                "Definitions cannot nest"
+                            raise Puppet::ParseError, msg
                         end
 
-                        @indefine = true
+                        @indefine = value
                     end
 
                     @last = value
@@ -295,4 +297,4 @@ module Puppet
     end
 end
 
-# $Id: lexer.rb 1726 2006-10-04 18:24:24Z luke $
+# $Id: lexer.rb 2393 2007-04-19 18:34:03Z luke $
